"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Resolver = void 0;
const node_path_1 = __importDefault(require("node:path"));
const debug_1 = __importDefault(require("debug"));
const typescript_1 = require("typescript");
const constants_1 = require("./constants");
const node_fs_1 = require("node:fs");
/* eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-call*/
const log = (0, debug_1.default)(constants_1.PROJECT_NAME);
function isParentDir(dir, parentToTest) {
    const diff = node_path_1.default.relative(parentToTest, dir);
    return diff.length === 0 || (!diff.startsWith('..') && !node_path_1.default.isAbsolute(diff));
}
class Resolver {
    constructor(ts, compilerOptions, alias) {
        this.ts = ts;
        this.compilerOptions = compilerOptions;
        this.cachedPackageJson = {};
        // Build matchers from compiler options
        this.matchers = [
            (moduleName, sourceFileName) => {
                const retVal = this.resolveModule(sourceFileName, moduleName);
                if (retVal) {
                    log(`${sourceFileName}: Rewrite module ${moduleName} to ${retVal}`);
                    return retVal;
                }
                return null;
            },
        ];
        if (alias) {
            // Build matchers from alias property of the plugin
            this.matchers.push(...Object.entries(alias).map(([from, to]) => {
                const fromRegex = new RegExp(from);
                return function (moduleName, sourceFileName) {
                    if (fromRegex.test(moduleName)) {
                        const retVal = moduleName.replace(fromRegex, to);
                        log(`${sourceFileName}: Rewrite module ${moduleName} to ${retVal} with regex /${from}/ => ${to}`);
                        return retVal;
                    }
                    return null;
                };
            }));
        }
        this.matchers.push();
    }
    resolve(moduleName, sourceFileName) {
        let retVal = null;
        this.matchers.forEach(function (matcher) {
            const result = matcher(moduleName, sourceFileName);
            if (result != null) {
                retVal = result;
            }
        });
        return retVal;
    }
    resolveModule(sourceFilePath, moduleName) {
        const { resolvedModule } = this.ts.resolveModuleName(moduleName, sourceFilePath, this.compilerOptions, this.ts.sys);
        if (!resolvedModule || resolvedModule.packageId) {
            return null;
        }
        let resolvedFileDir = node_path_1.default.dirname(resolvedModule.resolvedFileName);
        const resolvedFileExt = resolvedModule.extension;
        const resolvedFileBaseWithoutExt = resolvedModule.resolvedFileName.substring(resolvedFileDir.length + 1, resolvedModule.resolvedFileName.length - resolvedFileExt.length);
        const originalSourceFileDir = node_path_1.default.dirname(sourceFilePath);
        let sourceFileDir = originalSourceFileDir;
        if (this.compilerOptions.rootDirs) {
            let fileRootDir;
            let moduleRootDir;
            this.compilerOptions.rootDirs.forEach(function (rootDir) {
                if (isParentDir(resolvedFileDir, rootDir) && (!moduleRootDir || rootDir.length > moduleRootDir.length)) {
                    moduleRootDir = rootDir;
                }
                if (isParentDir(sourceFileDir, rootDir) && (!fileRootDir || rootDir.length > fileRootDir.length)) {
                    fileRootDir = rootDir;
                }
            });
            if (fileRootDir && moduleRootDir) {
                sourceFileDir = node_path_1.default.relative(fileRootDir, sourceFileDir);
                resolvedFileDir = node_path_1.default.relative(moduleRootDir, resolvedFileDir);
            }
        }
        const normalisedResolvedFileDir = node_path_1.default.relative(sourceFileDir, resolvedFileDir);
        let normalisedResolvedFileBase = resolvedFileBaseWithoutExt;
        if (this.isESM(originalSourceFileDir)) {
            if (resolvedFileExt === '.ts' || resolvedFileExt === '.d.ts') {
                normalisedResolvedFileBase += '.js';
            }
            else if (resolvedFileExt === '.mts') {
                normalisedResolvedFileBase += '.mjs';
            }
            else {
                normalisedResolvedFileBase += resolvedFileExt;
            }
        }
        let normalisedResolvedFilePath = node_path_1.default.join(normalisedResolvedFileDir, normalisedResolvedFileBase);
        if (!normalisedResolvedFilePath.startsWith('.')) {
            normalisedResolvedFilePath = `./${normalisedResolvedFilePath}`;
        }
        return normalisedResolvedFilePath;
    }
    isESM(sourceFileDir) {
        if (this.compilerOptions.module === typescript_1.ModuleKind.Node16 || this.compilerOptions.module === typescript_1.ModuleKind.NodeNext) {
            const pkg = this.findNearestPackageJson(sourceFileDir);
            return pkg && pkg.found && pkg.type === 'module';
        }
        return (this.compilerOptions.module !== typescript_1.ModuleKind.None &&
            this.compilerOptions.module !== typescript_1.ModuleKind.CommonJS &&
            this.compilerOptions.module !== typescript_1.ModuleKind.AMD &&
            this.compilerOptions.module !== typescript_1.ModuleKind.UMD &&
            this.compilerOptions.module !== typescript_1.ModuleKind.System);
    }
    findNearestPackageJson(srcDir) {
        if (!this.cachedPackageJson[srcDir]) {
            let oldSrcDir = srcDir;
            let currenSrcDir = srcDir;
            while (!this.cachedPackageJson[srcDir]) {
                // Try to read the package.json
                const pkgJsonPath = node_path_1.default.resolve(currenSrcDir, 'package.json');
                try {
                    if ((0, node_fs_1.existsSync)(pkgJsonPath)) {
                        const data = JSON.parse((0, node_fs_1.readFileSync)(pkgJsonPath, 'utf-8'));
                        if (typeof data === 'object' &&
                            (data.type == null || data.type.toLowerCase() === 'commonjs' || data.type.toLowerCase() === 'module')) {
                            this.cachedPackageJson[srcDir] = this.cachedPackageJson[currenSrcDir] = {
                                found: true,
                                type: data.type ? data.type.toLowerCase() : 'commonjs',
                            };
                        }
                    }
                }
                catch (err) {
                    // Ignore error
                }
                if (!this.cachedPackageJson[srcDir]) {
                    // Go to parent directory
                    oldSrcDir = currenSrcDir;
                    currenSrcDir = node_path_1.default.resolve(currenSrcDir, '..');
                    if (this.cachedPackageJson[currenSrcDir]) {
                        // In the cache
                        this.cachedPackageJson[srcDir] = this.cachedPackageJson[currenSrcDir];
                        break;
                    }
                    if (currenSrcDir === oldSrcDir) {
                        // Cannot go any further
                        this.cachedPackageJson[srcDir] = this.cachedPackageJson[currenSrcDir] = {
                            found: false,
                        };
                        break;
                    }
                }
            }
        }
        return this.cachedPackageJson[srcDir];
    }
}
exports.Resolver = Resolver;
//# sourceMappingURL=resolver.js.map